
/*
  Rev 13   of An ESP8266 Cheap and Simple Wifi Shield for Arduino and other micros
  Load this sketch to the ESP8266-01 module or the Adafruit HAZZAH ESP8266 module.

  The ESP8266 can only handle sending one TCP packet (of 1460 bytes of data) at a time.
  This code and the associated pfodESP8266WiFi non-blocking library allows incoming Serial data to be read and buffered while the ESP is handling the
  previous packet, waiting for it to be ACKed.  This allows data to be streamed at higher baud rates.

  Window's OS takes about 200mS to ACK a single TCP packet. Since the ESP8266 blocks handling more outgoing data until the last packet is ACKed,
  so, even with this non-blocking code and library, for continual data transmission you should limit the Serial baud rate to <= 57600 baud to avoid loss of data.

  For linux and Android, the ACKs come back much faster, 10-40mS, so when connecting from these OS's you can send data continually at 115200 baud.

  If the WiFi connection is poor and there are re-transmissions of lost packets, it takes longer for the ESP to successfully send the packed and
  you will need to set a lower baud rate to avoid lost data.
  19200 is the maximum practical baud rate for communicating with Arduino UNO and Arduino Mega boards

  This code could be extended to have multiple input buffers to handle higher burst rates of data. This is left as an exercise for the reader.

  This code also has a connection idle timeout.  If there is no incoming data from the connected client, this code will close the connection after
  the connection timeout time set in webpage config (default 15 secs).  The prevents half-closed connections from locking out new connections.  If you set
  connection timeout to 0 the connection will never time out, not recommended.

  This code also batches outgoing data by delaying sending of the buffered data until either the buffer is full, a full TCP packet, or no data has
  been received from the Serial port for SEND_DELAY_TIME (10mS).  This improves through put and allows higher baud rates without data loss.

  Rev 13 fix compile failure introduced by Arduino 1.6.6
  Rev 12 minor changes to handling new clients
  Rev 11 handles and closes other connections if already connected, set timeout as set from config was always 0 in rev10
  Rev 10 adds non-blocking library, pfodESP8266WiFi, and added baud rate setting to web config page
  Rev 8 removes 200mS wait after client.write()  client.write() is a blocking call.
  Rev 7 copies one byte at a time to Serial tx. Avoids calling client.flush() and waits 200mS after calling client.write()
  Rev 6 checked tx available
  Rev 5 added WiFiClient.stopAll() to clean up memory on disconnect
  Rev 4 added option connection disconnection messsages use #define CONNECTION_MESSAGES to enable them
  Rev 3 added connection timeout
  Rev 4 change tcpNoDelay to true clients and only monitors incoming data to reset connection timer.

  Load the Arduino with an empty sketch,
  connect USB to Serial TX to Flash TX
  connect USB to Serial RX to Flash RX
  short out the Flash Link and apply power to get into flash mode

  To Config
  Short the config link and apply power
    and connect to the pfodWifiWebConfig access point
  using the password set in  #define pfodWifiWebConfigPASSWORD  (below)
  Scan this password from the QR code generated with http://www.forward.com.au/pfod/secureChallengeResponse/keyGenerator/index.html

  Once you connect to the Access Point open the webpage
  http://10.1.1.1 to setup your network parameters
  ssid, password and either DHCP or staticIP and portNo.
  Then when you reboot (with the config link removed) you can connect to the ip:portNo and pass data via the UART in both directions.

  The micro board that is connected to this module should have setup() code stating with

  void setup() {
   delay(1000);  // skip skip the ESP8266 debug output
   Serial.begin(19200); // or whatever baud rate you configured via the webpage

    .... rest of setup here
     ..


  see http://www.forward.com.au/pfod/pfodWifiConfig/ESP8266/pfodWifiConfig_ESP8266.html for details
  For an example QR code image look in the directory this file is in.
*/

/**
    Cheap and Simple Wifi Shield for Arduino and other micros
   http://www.forward.com.au/pfod/CheapWifiSheild/index.html

   (c)2015 Forward Computing and Control Pty. Ltd.
   This code may be freely used for both private and commerical use.
   Provide this copyright is maintained.
*/

/**
  Code modified to create an easy to config serial to WiFi device for posting to ThingSpeak from an ESP8266.
  To use: create a serial connection at 9600 baud. Send fully formed ThingSpeak data string with a period no
  shorter than 15 seconds (ThingSpeak posting limit). A ending '~' determines the end of the string and
  triggers the ESP8266 to post. Note, a 30 second posting interval works well.

  Example, send 100.0 to ThingSpeak field1: field1=100.0~

  Note: debugging information has been left in and will return the following results with the above post viewed in a serial monitor:
  field1=100.0
  Client not connected and data ready
  Waiting for connection
  Connected and Data posted
  RSSI:-66
  
   (c) Steven Mann 2016
   This code may be freely used for both private and commerical use.
   Provide this copyright is maintained.
*/

#include <EEPROM.h>
#include <pfodESP8266Utils.h>
#include <pfodESP8266WiFi.h>
#include <pfodWiFiClient.h>
#include <pfodESP8266WebServer.h>
#include <pfodESP8266BufferedClient.h>

//#include <ESP8266WiFi.h>

//---------- Arduin0 1.6.6+ needs these declarations -------------
void handleNotFound();
void handleRoot();
void handleConfig();
void setupAP(const char* ssid_wifi, const char* password_wifi);
void closeConnection();


// normally DEBUG is commented out
//#define DEBUG
// Initial baud rate for DEBUG is 115200,
// If not starting in web config mode the baud rate is switched to the configured baud.

// uncomment this to connect / disconnect messages on Serial out.
//#define CONNECTION_MESSAGES

// the default to show in the config web page
const int DEFAULT_CONNECTION_TIMEOUT_SEC = 15;
WiFiServer server(80);
WiFiClient client; // just one client reused
pfodESP8266BufferedClient bufferedClient;

// =============== start of pfodWifiWebConfig settings ==============
// update this define with the password from your QR code
//  http://www.forward.com.au/pfod/secureChallengeResponse/keyGenerator/index.html
#define pfodWifiWebConfigPASSWORD "PASS#1234#"
#define pfodWifiWebConfigAP "ThingSpeak Client"

// note pfodSecurity uses 19 bytes of eeprom usually starting from 0 so
// start the eeprom address from 20 for configureWifiConfig
const uint8_t webConfigEEPROMStartAddress = 20;

int wifiSetup_pin = 2; // name the input pin for setup mode detection GPIO2 on most ESP8266 boards
// =============== end of pfodWifiWebConfig settings ==============

// On ESP8266-01 and Adafruit HAZZAH ESP8266, connect LED + 270ohm resistor from D2 (GPIO2) to +3V3 to indicate when in config mode

// set the EEPROM structure
struct EEPROM_storage {
  uint32_t timeout;
  uint32_t baudRate;
  uint16_t portNo;
  char ssid[pfodESP8266Utils::MAX_SSID_LEN + 1]; // WIFI ssid + null
  char password[pfodESP8266Utils::MAX_PASSWORD_LEN + 1]; // WiFi password,  if empyt use OPEN, else use AUTO (WEP/WPA/WPA2) + null
  char staticIP[pfodESP8266Utils::MAX_STATICIP_LEN + 1]; // staticIP, if empty use DHCP + null
  char ThingSpeakAPIKey[pfodESP8266Utils::MAX_SSID_LEN - 15]; // ThingSpeak API Key (16 characters)
} storage;
const int EEPROM_storageSize = sizeof(EEPROM_storage);

char ssid_found[pfodESP8266Utils::MAX_SSID_LEN + 1]; // max 32 chars + null
ESP8266WebServer webserver(80);  // this just sets portNo nothing else happens until begin() is called

byte inConfigMode = 0; // false
uint32_t timeout = 0;
const char *aps;
boolean SecondSetup = true;

//ThingSpeak Setup
byte ts_server[]  = {  184, 106, 153, 149 }; // IP Address for the ThingSpeak API
String writeAPIKey; // Holds the Write API Key in memory for ThingSpeak
WiFiClient ts_client; // Client for connecting to ThingSpeak server

// ThingSpeak Variable Setup
char thingSpeakAddress[] = "api.thingspeak.com";
boolean lastConnected = false;
///End ThingSpeak setup

//For Serial
String inputString = "";         // a string to hold incoming data
String InputStringPost = "";
boolean stringComplete = false;  // whether the string is complete
boolean stringPosted = true;
byte byteRead;


void setup ( void ) {
  // reserve 100 bytes for the inputString, data from other micro (connected via serial).
  inputString.reserve(100);
  InputStringPost.reserve(100);

  inConfigMode = 0; // not in config mode
  Serial.begin(9600); // Serial setup for comms to other micro
  EEPROM.begin(512); // must be greater than (wifiConfigStartAddress + EEPROM_storageSize)
  delay(10);
  pinMode(0, OUTPUT);
  digitalWrite(0, LOW); // make GPIO0 low after ESP8266 has initialized

  for (int i = 4; i > 0; i--) {
    delay(500);
  }

  if (digitalRead(wifiSetup_pin) == LOW) {
    inConfigMode = 1; // in config mode
    WiFi.mode(WIFI_AP_STA);
    // connect to temporary wifi network for setup
    // the features determine the format of the {set...} command
    setupAP(pfodWifiWebConfigAP, pfodWifiWebConfigPASSWORD);
  }
}


unsigned long timeoutTimerStart = 0;
const unsigned long SEND_DELAY_TIME = 10; // 10mS delay before sending buffer
unsigned long sendTimerStart = 0;
static const size_t SEND_BUFFER_SIZE = 1460; // Max data size for standard TCP/IP packet
static uint8_t sendBuffer[SEND_BUFFER_SIZE]; //
size_t sendBufferIdx = 0;
bool alreadyConnected = false;


// the loop routine runs over and over again forever:
void loop() {

  //Serial handling
  if (stringComplete) {
    Serial.println(inputString); // debugging
    InputStringPost = inputString;
    inputString = "";
    stringComplete = false;
    stringPosted = false;
  }
  serialEvent(); // Collect serial data

  if (inConfigMode) {

    if (inConfigMode) {
      webserver.handleClient();
      delay(0);
      return;
    }
    if (server.hasClient()) {   // avoid creating WiFiClient object if no connection,  ESP8266 specific
      WiFiClient anotherClient = server.available();
      if (!client) {
        client = anotherClient; // take this connection
      } else {
        anotherClient.stop(); // stop this extra connection
      }
    }  // anotherClient is disposed here at end of its block
    if (client) {
      // have client
      if (!client.connected()) {
        if (alreadyConnected) {
          // client closed so clean up
          closeConnection();
        }
      } else {
        // have connected client
        if (!alreadyConnected) {
          client.setNoDelay(true); // does not do much if anything
          bufferedClient.connect(&client); // buffer this client
          alreadyConnected = true;
          // start timer
          timeoutTimerStart = millis();
        }
      }
    }

    // NOTE: MUST call some bufferedClient method each loop to push out buffered data on delayed send timeout
    if (bufferedClient.connected()) {
      while ((bufferedClient.available() > 0) &&  (Serial.availableForWrite() > 0)) {
        // use Serial.availableForWrite to prevent loosing incoming data
        timeoutTimerStart = millis();  // reset timer if we have incoming data
        Serial.write(bufferedClient.read());
      }
    }

    // see if we should drop the connection
    if (alreadyConnected && (timeout > 0) && ((millis() - timeoutTimerStart) > timeout)) {
      closeConnection();
    }
  }

  /// Other normal code
  if (SecondSetup == true) {
    SecondSetup = false; // Only Run once
    // else button was not pressed continue to load the stored network settings
    // else use configured setttings from EEPROM
    uint8_t * byteStorageRead = (uint8_t *)&storage;
    for (size_t i = 0; i < EEPROM_storageSize; i++) {
      byteStorageRead[i] = EEPROM.read(webConfigEEPROMStartAddress + i);
    }
    timeout = storage.timeout; // set the timeout

    WiFi.begin(storage.ssid, storage.password);
    WiFi.mode(WIFI_STA);  // Station Mode only (no access point)
    while (WiFi.status() != WL_CONNECTED) {
      delay(500);
      Serial.print(".");
    }
    writeAPIKey = storage.ThingSpeakAPIKey;    // Write API Key for a ThingSpeak Channel
  }
  DealWithThingSpeak();
}


void serialEvent() {
  while (Serial.available()) {
    // get the new byte:
    char inChar = (char)Serial.read();
    // add it to the inputString if valid
    if ((inChar > 34) && (inChar < 122)) {
      inputString += inChar;
    }
    // if the incoming character is a '~" then its the end of the data stream and time to post
    if (inChar == '~') {
      stringComplete = true;
    }
  }
}


void DealWithThingSpeak()
{
  // This line is required
  ts_client.read();

  // Disconnect from ThingSpeak
  if (!ts_client.connected() && lastConnected)
  {
    ts_client.stop();
    Serial.println ( "Stopping the client" );
  }

  ts_client.read();
  // Update ThingSpeak
  if (!ts_client.connected() && !stringPosted)
  {
    Serial.println ( "Client not connected and data ready" );
    updateThingSpeak(InputStringPost);
    // stringPosted = true;
  }
  lastConnected = ts_client.connected();
}


void updateThingSpeak(String tsData)
{
  Serial.println ( "Waiting for connection" );
  if (ts_client.connect(thingSpeakAddress, 80))
  {
    ts_client.print("POST /update HTTP/1.1\n");
    ts_client.print("Host: api.thingspeak.com\n");
    ts_client.print("Connection: close\n");
    ts_client.print("X-THINGSPEAKAPIKEY: " + writeAPIKey + "\n");
    ts_client.print("Content-Type: application/x-www-form-urlencoded\n");
    ts_client.print("Content-Length: ");
    ts_client.print(tsData.length());
    ts_client.print("\n\n");

    ts_client.print(tsData);
    stringPosted = true;
    Serial.println ( "Connected and Data posted" );

    // print the received signal strength:
    long rssi = WiFi.RSSI();
    Serial.print("RSSI:");
    Serial.println(rssi);

  }
}


void closeConnection() {
#ifdef CONNECTION_MESSAGES
  Serial.println("ConnectionClosed");
#endif
  alreadyConnected = false;
  bufferedClient.stop(); // clears client reference
  client.stop(); // stop client and tcp buffer.
  if (server.hasClient()) {   // avoid creating WiFiClient object if no connection,  ESP8266 specific
    client = server.available();
  } // else just keep client that was just stopped will evaluate to false
}


void setupAP(const char* ssid_wifi, const char* password_wifi) {
  aps = pfodESP8266Utils::scanForStrongestAP((char*)&ssid_found, pfodESP8266Utils::MAX_SSID_LEN + 1);
  delay(0);
  IPAddress local_ip = IPAddress(10, 1, 1, 1);
  IPAddress gateway_ip = IPAddress(10, 1, 1, 1);
  IPAddress subnet_ip = IPAddress(255, 255, 255, 0);
#ifdef DEBUG
  Serial.println(F("configure pfodWifiWebConfig"));
#endif
  WiFi.softAP(ssid_wifi, password_wifi);

#ifdef DEBUG
  Serial.println(F("Access Point setup"));
#endif
  WiFi.softAPConfig(local_ip, gateway_ip, subnet_ip);

#ifdef DEBUG
  Serial.println("done");
  IPAddress myIP = WiFi.softAPIP();
  Serial.print(F("AP IP address: "));
  Serial.println(myIP);
#endif
  delay(10);
  webserver.on ( "/", handleRoot );
  webserver.on ( "/config", handleConfig );
  webserver.onNotFound ( handleNotFound );
  webserver.begin();
#ifdef DEBUG
  Serial.println ( "HTTP webserver started" );
#endif
}

void handleConfig() {
  // set defaults
  uint16_t portNo = 80;
  uint32_t timeout = DEFAULT_CONNECTION_TIMEOUT_SEC * 1000; // time out in 15 sec
  uint32_t baudRate = 9600;

  if (webserver.args() > 0) {
#ifdef DEBUG
    String message = "Config results\n\n";
    message += "URI: ";
    message += webserver.uri();
    message += "\nMethod: ";
    message += ( webserver.method() == HTTP_GET ) ? "GET" : "POST";
    message += "\nArguments: ";
    message += webserver.args();
    message += "\n";
    for ( uint8_t i = 0; i < webserver.args(); i++ ) {
      message += " " + webserver.argName ( i ) + ": " + webserver.arg ( i ) + "\n";
    }
    Serial.println(message);
#endif

    uint8_t numOfArgs = webserver.args();
    const char *strPtr;
    uint8_t i = 0;
    for (; (i < numOfArgs); i++ ) {
      // check field numbers
      if (webserver.argName(i)[0] == '1') {
        pfodESP8266Utils::strncpy_safe(storage.ssid, (webserver.arg(i)).c_str(), pfodESP8266Utils::MAX_SSID_LEN);
        pfodESP8266Utils::urldecode2(storage.ssid, storage.ssid); // result is always <= source so just copy over
      } else if (webserver.argName(i)[0] == '2') {
        pfodESP8266Utils::strncpy_safe(storage.password, (webserver.arg(i)).c_str(), pfodESP8266Utils::MAX_PASSWORD_LEN);
        pfodESP8266Utils::urldecode2(storage.password, storage.password); // result is always <= source so just copy over
        // if password all blanks make it empty
        if (pfodESP8266Utils::isEmpty(storage.password)) {
          storage.password[0] = '\0';
        }
      }
      else if (webserver.argName(i)[0] == '7') { /// ADDED for thing Speak Key
        pfodESP8266Utils::strncpy_safe(storage.ThingSpeakAPIKey, (webserver.arg(i)).c_str(), pfodESP8266Utils::MAX_SSID_LEN);///16+1); ///
        pfodESP8266Utils::urldecode2(storage.ThingSpeakAPIKey, storage.ThingSpeakAPIKey); // result is always <= source so just copy over ///
        // if password all blanks make it empty
        if (pfodESP8266Utils::isEmpty(storage.ThingSpeakAPIKey)) {
          storage.ThingSpeakAPIKey[0] = '\0';
        }
      }
    }

    uint8_t * byteStorage = (uint8_t *)&storage;
    for (size_t i = 0; i < EEPROM_storageSize; i++) {
      EEPROM.write(webConfigEEPROMStartAddress + i, byteStorage[i]);
    }
    delay(0);
    EEPROM.commit();
  } // else if no args just return current settings

  delay(0);
  struct EEPROM_storage storageRead;
  uint8_t * byteStorageRead = (uint8_t *)&storageRead;
  for (size_t i = 0; i < EEPROM_storageSize; i++) {
    byteStorageRead[i] = EEPROM.read(webConfigEEPROMStartAddress + i);
  }

  String rtnMsg = "<html>"
                  "<head>"
                  "<title>Setup</title>"
                  "<meta charset=\"utf-8\" />"
                  "<meta name=viewport content=\"width=device-width, initial-scale=1\">"
                  "</head>"
                  "<body>"
                  "<h2>Settings saved.</h2><br>Power cycle to connect to ";
  if (storageRead.password[0] == '\0') {
    rtnMsg += "the open network ";
  }
  rtnMsg += "<b>";
  rtnMsg += storageRead.ssid;
  rtnMsg += "</b>";

  if (storageRead.staticIP[0] == '\0') {
    rtnMsg += "<br> using DCHP to get its IP address";
  } else { // staticIP
    rtnMsg += "<br> using IP addess ";
    rtnMsg += "<b>";
    rtnMsg += storageRead.staticIP;
    rtnMsg += "</b>";
  }
  rtnMsg += "<br><br>ThingSpeak write API Key: ";
  rtnMsg += storageRead.ThingSpeakAPIKey;
  "</body>"
  "</html>";

  webserver.send ( 200, "text/html", rtnMsg );
}


void handleRoot() {
  String msg;
  msg = "<html>"
        "<head>"
        "<title>Setup</title>"
        "<meta charset=\"utf-8\" />"
        "<meta name=viewport content=\"width=device-width, initial-scale=1\">"
        "</head>"
        "<body>"
        "<h2>Setup</h2>"
        "<p>Use this form to configure your Wifi network.</p>"
        "<form class=\"form\" method=\"post\" action=\"/config\" >"
        "<p class=\"name\">"
        "<label for=\"name\">Network SSID</label><br>"
        "(Displays strongest accesspoint, else enter your accesspoint name)<br>"
        "<input type=\"text\" name=\"1\" id=\"ssid\" placeholder=\"wifi network name\"  required "; // field 1

  if (*aps != '\0') {
    msg += " value=\"";
    msg += aps;
    msg += "\" ";
  }
  msg += " />"
         "<p class=\"password\">"
         "<label for=\"password\">Password for WEP/WPA/WPA2 (enter a space if there is no password, i.e. OPEN)</label><br>"
         "<input type=\"text\" name=\"2\" id=\"password\" placeholder=\"wifi network password\" autocomplete=\"off\" required " // field 2
         "</p>"
         "<p class=\"static_ip\">"
         "<label for=\"static_ip\">Set the Static IP for this device</label><br>"
         "(If this field is empty, DHCP will be used to get an IP address)<br>"
         "<input type=\"text\" name=\"3\" id=\"static_ip\" placeholder=\"192.168.4.99\" "  // field 3
         " pattern=\"\\b(?:(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\\.){3}(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\\b\"/>"
         "</p>"
         "<p class=\"ThingSpeakAPIKey\">"
         "<label for=\"ThingSpeakAPIKey\">ThingSpeak write API Key</label><br>"
         "<input type=\"text\" name=\"7\" id=\"ThingSpeakAPIKey\" placeholder=\"ThingSpeak write API Key\" autocomplete=\"off\" required " // field 2
         "</p>"
         "<p class=\"submit\">"
         "<input type=\"submit\" value=\"Configure\"  />"
         "</p>"
         "</form>"
         "</body>"
         "</html>";

  webserver.send ( 200, "text/html", msg );
}


void handleNotFound() {
  handleRoot();
}
